/* $Id$ */
#include <lcms.h>
#include "common.h"
#include "cmsprofile.h"
#include "cmsimage.h"
#include "module.h"

struct cms_profile {
  cmsHPROFILE profile;
};

static cmsHPROFILE get_profile_CIExyY(LPcmsCIExyY white, LPcmsCIExyYTRIPLE primaries, double gamma);
static CmsProfile *cmsProfile_alloc(void);
static void cmsProfile_free(CmsProfile *profile);

static DWORD image_dw_type(Image *image);
static void transform_data(cmsHTRANSFORM transform, Image *image_in, Image *image_out);
static void transform_indexed_colors(cmsHTRANSFORM transform, Image *image_in, Image *image_out);
#if defined DEBUG
static void info_print(cmsHPROFILE profile);
#endif /* DEBUG */

CmsProfile *cmsProfile_open_mem(void *ptr, size_t len)
{
  CmsProfile *profile;
  DEBUG_PRINT("open_mem()\n");
  profile = cmsProfile_alloc();
  if (ptr != NULL) {
    profile->profile = cmsOpenProfileFromMem(ptr, len);
#if defined DEBUG
    info_print(profile->profile);
#endif /* DEBUG */
  }
  return profile;
}

CmsProfile *cmsProfile_open_file(const char *filename)
{
  CmsProfile *profile;
  DEBUG_PRINT("open_file() filename = '%s'\n", filename);
  profile = cmsProfile_alloc();
  if (filename != NULL) {
    profile->profile = cmsOpenProfileFromFile(filename, "r");
#if defined DEBUG
    info_print(profile->profile);
#endif /* DEBUG */
  }
  return profile;
}

CmsProfile *cmsProfile_open_sRGB(void)
{
  CmsProfile *profile;
  DEBUG_PRINT("open_sRGB()\n");
  profile = cmsProfile_alloc();
  profile->profile = cmsCreate_sRGBProfile();
#if defined DEBUG
  info_print(profile->profile);
#endif /* DEBUG */
  return profile;
}

CmsProfile *cmsProfile_open_chromaticities(double white_x, double white_y,
                                           double red_x, double red_y,
                                           double green_x, double green_y,
                                           double blue_x, double blue_y,
                                           double gamma)
{
  CmsProfile *profile;
  cmsCIExyY white;
  cmsCIExyYTRIPLE primaries;
  DEBUG_PRINT("open_chromaticities()\n");
  profile = cmsProfile_alloc();
  white.x = white_x;
  white.y = white_y;
  primaries.Red.x = red_x;
  primaries.Red.y = red_y;
  primaries.Green.x = green_x;
  primaries.Green.y = green_y;
  primaries.Blue.x = blue_x;
  primaries.Blue.y = blue_y;
  white.Y = primaries.Red.Y = primaries.Green.Y = primaries.Blue.Y = 1.0;
  profile->profile = get_profile_CIExyY(&white, &primaries, gamma);
#if defined DEBUG
  info_print(profile->profile);
#endif /* DEBUG */
  return profile;
}

CmsProfile *cmsProfile_open_D65(double gamma)
{
  CmsProfile *profile;
  cmsCIExyY white = { 0.3127, 0.3290, 1.0 }; /* D65 */
  cmsCIExyYTRIPLE primaries = {
    { 0.6400, 0.3300, 1.0 },
    { 0.3000, 0.6000, 1.0 },
    { 0.1500, 0.0600, 1.0 }
  };
  profile = cmsProfile_alloc();
  profile->profile = get_profile_CIExyY(&white, &primaries, gamma);
#if defined DEBUG
  info_print(profile->profile);
#endif /* DEBUG */
  return profile;
}

int cmsProfile_close(CmsProfile *profile)
{
  int rc = FALSE;
  if (profile == NULL) {
    return rc;
  }
  if (cmsCloseProfile(profile->profile)) {
    profile->profile = NULL;
    rc = TRUE;
  }
  cmsProfile_free(profile);
  return rc;
}

static cmsHPROFILE get_profile_CIExyY(LPcmsCIExyY white, LPcmsCIExyYTRIPLE primaries, double gamma)
{
  cmsHPROFILE profile;
  LPGAMMATABLE trans[3]; /* gamma_table */

  trans[0] = trans[1] = trans[2] = cmsBuildGamma(256, 1.0/gamma);
  if (trans[0] == NULL) {
    return NULL;
  }
  profile = cmsCreateRGBProfile(white, primaries, trans);
  cmsFreeGamma(trans[0]);
  return profile;
}

static CmsProfile *cmsProfile_alloc(void)
{
  CmsProfile *p;
  p = XMALLOC(CmsProfile, 1);
  p->profile = NULL;
  return p;
}

static void cmsProfile_free(CmsProfile *profile)
{
  XFREE(profile);
}


Image *image_transform_new(Image *image_in, CmsProfile *profile_in, CmsProfile *profile_out)
{
  Image *image_out;
  unsigned char *data;
  ImageType type;
  size_t size;
  int w, h, xoff, yoff;

  image_get_size(image_in, &w, &h);
  image_get_offset(image_in, &xoff, &yoff);
  type = image_get_type(image_in);
  size = w*h*imageType_bytes_per_pixel(type);
  data = XMALLOC(unsigned char, size);
  image_out = image_new();
  image_set_type(image_out, type);
  image_set_data(image_out, data);
  image_set_size(image_out, w, h);
  image_set_offset(image_out, xoff, yoff);
  if (type == IMAGE_TYPE_INDEXED) {
    image_transform_pallete(image_in, profile_in, image_out, profile_out);
  } else {
    image_transform_data(image_in, profile_in, image_out, profile_out);
  }
  return image_out;
}

void image_transform_data(Image *image_in, CmsProfile *profile_in, Image *image_out, CmsProfile *profile_out)
{
  cmsHPROFILE pin, pout;
  cmsHTRANSFORM transform;
  int intent = INTENT_PERCEPTUAL;
  DWORD dw_flags = 0;
  DWORD dw_type_in, dw_type_out;

  pin = profile_in->profile;
  pout = profile_out->profile;
  dw_type_in = image_dw_type(image_in);
  dw_type_out = image_dw_type(image_out);
  transform = cmsCreateTransform(pin, dw_type_in, pout, dw_type_out,
                                 intent, dw_flags);
  transform_data(transform, image_in, image_out);
  cmsDeleteTransform(transform);
}

void image_transform_pallete(Image *image_in, CmsProfile *profile_in, Image *image_out, CmsProfile *profile_out)
{
  cmsHPROFILE pin, pout;
  cmsHTRANSFORM transform;
  int intent = INTENT_PERCEPTUAL;
  DWORD dw_flags = 0;
  DWORD dw_type_in, dw_type_out;

  pin = profile_in->profile;
  pout = profile_out->profile;
  dw_type_in = image_dw_type(image_in);
  dw_type_out = image_dw_type(image_out);
  DEBUG_PRINT("type_in=%d, type_out=%d\n",
              image_get_type(image_in), image_get_type(image_out));
  transform = cmsCreateTransform(pin, dw_type_in, pout, dw_type_out,
                                 intent, dw_flags);
  transform_indexed_colors(transform, image_in, image_out);
  cmsDeleteTransform(transform);
}

static DWORD image_dw_type(Image *image)
{
  DWORD dw_type;
  ImageType type;

  type = image_get_type(image);
  switch (type) {
  case IMAGE_TYPE_INDEXED:
    dw_type = TYPE_RGBA_8;
    break;
  case IMAGE_TYPE_RGB_8:
    dw_type = TYPE_RGB_8;
    break;
  case IMAGE_TYPE_RGBA_8:
    dw_type = TYPE_RGBA_8;
    break;
  case IMAGE_TYPE_RGB_16:
    dw_type = TYPE_RGB_16;
    break;
  case IMAGE_TYPE_RGBA_16:
    dw_type = TYPE_RGBA_16;
    break;
  default:
    break;
  }
  return dw_type;
}

static void transform_data(cmsHTRANSFORM transform, Image *image_in, Image *image_out)
{
  unsigned char *data_in, *data_out;
  unsigned char *p_in, *p_out;
  int bpp_in, bpp_out;
  int w_in, h_in, w_out, h_out;
  int xoff_in, yoff_in, xoff_out, yoff_out;

  DEBUG_PRINT("transform_data() --- start.\n");
  image_get_size(image_in, &w_in, &h_in);
  image_get_offset(image_in, &xoff_in, &yoff_in);
  bpp_in = imageType_bytes_per_pixel(image_get_type(image_in));
  if ((bpp_in <= 0) || (w_in <= 0) || (h_in <= 0)) {
    DEBUG_PRINT("bpp_in = %d, w_in = %d, h_in = %d\n", bpp_in, w_in, h_in);
    return;
  }
  DEBUG_PRINT("transform_data(): in check OK.\n");
  image_get_size(image_out, &w_out, &h_out);
  image_get_offset(image_out, &xoff_out, &yoff_out);
  bpp_out = imageType_bytes_per_pixel(image_get_type(image_out));
  if ((bpp_out <= 0) || (w_out <= 0) || (h_out <= 0)) {
    DEBUG_PRINT("bpp_out = %d, w_out = %d, h_out = %d\n", bpp_out, w_out, h_out);
    return;
  }
  DEBUG_PRINT("transform_data(): out check OK.\n");
  if ((w_in > w_out) || (h_in > h_out)) {
    DEBUG_PRINT("w_in = %d, w_out = %d, h_in = %d, h_out = %d\n", w_in, w_out, h_in, h_out);
    return;
  }
  DEBUG_PRINT("transform_data(): check OK.\n");
  data_in = image_get_data(image_in);
  data_out = image_get_data(image_out);
  p_in = data_in;
  p_out = data_out+(w_out*yoff_in+xoff_in)*bpp_out;
  if ((xoff_in == 0) && (xoff_out == 0) && (w_in == w_out)) {
    cmsDoTransform(transform, p_in, p_out, w_in*h_in);
  } else {
    for (int j = 0; j < h_in; j++) {
      cmsDoTransform(transform, p_in, p_out, w_in);
      p_in += w_in*bpp_in;
      p_out += w_out*bpp_out;
    }
  }
  DEBUG_PRINT("transform_data() --- end.\n");
}

static void transform_indexed_colors(cmsHTRANSFORM transform, Image *image_in, Image *image_out)
{
  int n;

  n = image_get_num_colors(image_in);
  image_set_num_colors(image_out, n);
  for (int i = 0; i < n; i++) {
    unsigned char din[4];
    unsigned char dout[4];

    image_get_color(image_in, i, &din[0], &din[1], &din[2], &din[3]);
    dout[3] = din[3]; /* copy alpha channel */
    cmsDoTransform(transform, din, dout, 1);
    image_set_color(image_out, i, dout[0], dout[1], dout[2], dout[3]);
  }
}

#if defined DEBUG
static void info_print(cmsHPROFILE profile)
{
  FILE *fp;

  fp = stderr;
#if 0
  fprintf(fp, "product name: %s\n", cmsTakeProductName(profile));
  fprintf(fp, "product desc: %s\n", cmsTakeProductDesc(profile));
  fprintf(fp, "manufacturer: %s\n", cmsTakeManufacturer(profile));
  fprintf(fp, "model: %s\n", cmsTakeModel(profile));
  fprintf(fp, "copyright: %s\n", cmsTakeCopyright(profile));
  fprintf(fp, "ICC version: %x\n", cmsGetProfileICCversion(profile));
  fprintf(fp, "ICC header flags: %x\n", cmsTakeHeaderFlags(profile));
  fprintf(fp, "ICC header attributes: %x\n", cmsTakeHeaderAttributes(profile));
  fprintf(fp, "intent field of the header: %x\n", cmsTakeRenderingIntent(profile));
  fprintf(fp, "number of tags: %d\n", cmsGetTagCount(profile));
  fprintf(fp, "\n");
#endif
}
#endif /* DEBUG */
