/* $Id$ */
#include "common.h"
#include "module.h"
#include "image_utils.h"
#include "cmsprofile.h"
#include "cmsimage.h"

#define NUM_FRAMES 3

struct imageWork {
  ImageType type;
  Image *current;
  Image *previous;
  Image *frame[NUM_FRAMES];
  int num_frames;
  int last_index;
  BgPrefer bg_prefer;
  int bgindex;
  unsigned char *bg;
  unsigned char *bguser;
  int dispose_op;
  int updated_xoff;
  int updated_yoff;
  int updated_width;
  int updated_height;
  CmsProfile *profile_embedded;
  CmsProfile *profile_in;
  CmsProfile *profile_out;
  int use_infile;
};

static ImageType get_imageWork_type(ImageSet *images);
static CmsProfile *imageWork_get_profile_in(ImageWork *work);
static void image_copy_data(Image *dst, Image *src, ImageType type);
static void image_copy_pallete(Image *dst, Image *src);
static void image_fill_pixel(Image *image, ImageType type, unsigned char *pixel);
static void image_blend(ImageWork *work, Image *newframe, ImageType type,
                        const unsigned char *bg, int bgindex);
static void image_blend_index_to_rgb8(Image *image, Image *newframe, const unsigned char *bg, int bgindex);
static void image_blend_rgb(Image *image, Image *newframe, ImageType type, const unsigned char *bg);
static void blend_rgb8(unsigned char *p1, const unsigned char *p2, const unsigned char *bg);
static void blend_rgb16(unsigned char *p1, const unsigned char *p2, const unsigned char *bg);
static void get_rgb8(unsigned char *rgb, unsigned short *r, unsigned short *g, unsigned short *b);
static void set_rgb8(unsigned char *rgb, unsigned short r, unsigned short g, unsigned short b);
static void get_rgb16(unsigned char *rgb, unsigned short *r, unsigned short *g, unsigned short *b);
static void set_rgb16(unsigned char *rgb, unsigned short r, unsigned short g, unsigned short b);
static unsigned short get_c16(unsigned char *c16);
static void set_c16(unsigned char *c16, unsigned short c);
static unsigned short get_c8(unsigned char *c8);
static void set_c8(unsigned char *c8, unsigned short c);


void imageWork_init(ImageWork *work, ImageSet *images)
{
  ImageType type;
  size_t size;
  int width, height;

  type = get_imageWork_type(images);
  imageSet_get_size(images, &width, &height);
  size = imageType_bytes_per_pixel(type)*width*height;
  for (int i = 0; i < work->num_frames; i++) {
    Image *image;
    size_t size0;
    int w, h;
    image = work->frame[i];
    image_get_size(image, &w, &h);
    size0 = imageType_bytes_per_pixel(work->type)*w*h;
    if (size != size0) {
      unsigned char *data;
      data = image_get_data(image);
      if (data != NULL) {
        XFREE(data);
      }
      data = XMALLOC(unsigned char, size);
      image_set_data(image, data);
    }
    image_set_size(image, width, height);
    image_set_dispose_op(image, DISPOSE_OP_BACKGROUND);
  }
  work->type = type;
  work->current = work->frame[0];
  work->previous = NULL;
  work->last_index = 2;

  unsigned short r, g, b;
  imageSet_get_background_rgb(images, &r, &g, &b);
  DEBUG_PRINT("type = %d, bg = %d, %d, %d\n", type, r, g, b);
  switch (type) {
  case IMAGE_TYPE_INDEXED:
    work->bgindex = imageSet_get_background_index(images);
    work->bg_prefer = BG_INDEX;
    break;
  case IMAGE_TYPE_RGB_8:
  case IMAGE_TYPE_RGBA_8:
    set_rgb8(work->bg, r, g, b);
    work->bg_prefer = BG_RGB;
    break;
  case IMAGE_TYPE_RGB_16:
  case IMAGE_TYPE_RGBA_16:
    set_rgb16(work->bg, r, g, b);
    work->bg_prefer = BG_RGB;
    break;
  }
  work->profile_embedded = imageSet_get_profile(images);
#if defined DEBUG
  DEBUG_PRINT(" bg_prefer = %d\n", work->bg_prefer);
  DEBUG_PRINT(" bgindex = %d\n", work->bgindex);
  for (int i = 0; i < 6; i++) {
    DEBUG_PRINT(" bg[%d] = %x\n", i, work->bg[i]);
  }
  for (int i = 0; i < 6; i++) {
    DEBUG_PRINT(" bguser[%d] = %x\n", i, work->bguser[i]);
  }
#endif /* DEBUG */
}

void imageWork_get_background(ImageWork *work, unsigned short *r, unsigned short *g, unsigned short *b)
{
  BgPrefer bg_prefer;
  ImageType type;
  unsigned char *bg;

  bg_prefer = work->bg_prefer;
  type = work->type;
  switch (type) {
  case IMAGE_TYPE_INDEXED:
    if (bg_prefer == BG_USER) {
      bg = work->bguser;
      get_rgb8(bg, r, g, b);
    } else {
      unsigned char rgba[4];
#if defined DEBUG
      DEBUG_PRINT("num_colors=%d\n", image_get_num_colors(work->current));
      DEBUG_PRINT("work->bgindex=%d\n", work->bgindex);
      for (int i = 0; i < 6; i++) {
        DEBUG_PRINT(" bguser[%d] = %x\n", i, work->bguser[i]);
      }
#endif /* DEBUG */
      image_get_color(work->current, work->bgindex, &rgba[0], &rgba[1], &rgba[2], &rgba[3]);
      bg = rgba;
      if (rgba[3] < 0x80) {
        /* fully transparent */
        bg = work->bguser;
      }
      get_rgb8(bg, r, g, b);
    }
    break;
  case IMAGE_TYPE_RGB_8:
  case IMAGE_TYPE_RGBA_8:
    if (bg_prefer == BG_USER) {
      bg = work->bguser;
    } else {
      bg = work->bg;
    }
    get_rgb8(bg, r, g, b);
    break;
  case IMAGE_TYPE_RGB_16:
  case IMAGE_TYPE_RGBA_16:
    if (bg_prefer == BG_USER) {
      bg = work->bguser;
    } else {
      bg = work->bg;
    }
    get_rgb16(bg, r, g, b);
    break;
  default:
    DEBUG_PRINT("unknown image type %d. why?\n", type);
    break;
  }
  DEBUG_PRINT(" bg_prefer=%d type=%d (%d,%d,%d)\n", bg_prefer, type, r, g, b);
}

void imageWork_set_background(ImageWork *work, unsigned short r, unsigned short g, unsigned short b, BgPrefer prefer)
{
  work->bg_prefer = prefer;
  switch (work->type) {
  case IMAGE_TYPE_INDEXED:
  case IMAGE_TYPE_RGB_8:
  case IMAGE_TYPE_RGBA_8:
    set_rgb8(work->bguser, r, g, b);
    break;
  case IMAGE_TYPE_RGB_16:
  case IMAGE_TYPE_RGBA_16:
    set_rgb16(work->bguser, r, g, b);
    break;
  }
}

static ImageType get_imageWork_type(ImageSet *images)
{
  ImageType type;
  Image *image;
  int num;

  image = imageSet_get_first(images);
  type = image_get_type(image);
  DEBUG_PRINT("imageWork: type = %d\n", type);
  switch (type) {
  case IMAGE_TYPE_INDEXED:
    num = imageSet_get_num_frames(images);
    image = imageSet_get_first(images);
    for (int j = 0; j < num; j++) {
      int n = image_get_num_colors(image);
      for (int i = 0; i < n; i++) {
        unsigned char r, g, b, a;
        image_get_color(image, i, &r, &g, &b, &a);
        if ((a > 0x0) && (a < 0xff)) {
          type = IMAGE_TYPE_RGB_8;
          return type;
        }
      }
      image = image_next(image);
    }
    break;
  case IMAGE_TYPE_RGB_8:
  case IMAGE_TYPE_RGBA_8:
    type = IMAGE_TYPE_RGB_8;
    break;
  case IMAGE_TYPE_RGB_16:
  case IMAGE_TYPE_RGBA_16:
    type = IMAGE_TYPE_RGB_16;
    break;
  default:
    break;
  }
  return type;
}

ImageWork *imageWork_alloc(void)
{
  ImageWork *work;

  work = XMALLOC(ImageWork, 1);
  work->type = IMAGE_TYPE_UNKNOWN;
  work->num_frames = NUM_FRAMES;
  for (int i = 0; i < work->num_frames; i++) {
    Image *image;
    image = image_new();
    image_set_dispose_op(image, DISPOSE_OP_BACKGROUND);
    work->frame[i] = image;
  }
  work->current = work->frame[0];
  work->previous = NULL;
  work->last_index = 2;
  work->bg_prefer = BG_RGB;
  work->bgindex = -1;
  work->bg = XMALLOC(unsigned char, 6);
  work->bguser = XMALLOC(unsigned char, 6);
  set_rgb16(work->bg, 0x7f7f, 0x7f7f, 0x7f7f);
  set_rgb16(work->bguser, 0x7f7f, 0x7f7f, 0x7f7f);
  work->updated_xoff = 0;
  work->updated_yoff = 0;
  work->updated_width = 0;
  work->updated_height = 0;
  work->profile_embedded = NULL;
  work->profile_in = NULL;
  work->profile_out = NULL;
  work->use_infile = FALSE;
  return work;
}

void imageWork_free(ImageWork *work)
{
  for (int i = 0; i < work->num_frames; i++) {
    image_free(work->frame[i]);
  }
  XFREE(work);
}

void imageWork_update(ImageWork *work, Image *newframe)
{
  Image *image;
  Image *last_frame;
  ImageType newtype;
  int dispose_op0;
  int dispose_op;
  int blend_op;
  unsigned char *bg;
  int bgindex;
  int w, h, xoff, yoff;
  int full_image = FALSE;

  newtype = image_get_type(newframe);
  DEBUG_PRINT(" image type: work = %d / new = %d\n", work->type, newtype);
  last_frame = NULL;
  if (work->bg_prefer == BG_USER) {
    bg = work->bguser;
  } else {
    bg = work->bg;
  }
  bgindex = work->bgindex;
  image = work->current;
  image_set_type(image, work->type); /* FIXME */
  dispose_op = image_get_dispose_op(newframe);
  if (dispose_op == DISPOSE_OP_PREVIOUS) {
    work->last_index = (work->last_index == 2) ? 1 : 2;
    last_frame = work->frame[work->last_index];
    image_copy_data(last_frame, image, work->type);
  }
  /* dispose operation */
  dispose_op0 = image_get_dispose_op(image);
  DEBUG_PRINT(" dispose_op type: %d\n", dispose_op0);
  switch (dispose_op0) {
  case DISPOSE_OP_NONE:
    break;
  case DISPOSE_OP_BACKGROUND:
    if (work->bg_prefer == BG_INDEX) {
      unsigned char bgi;
      bgi = (unsigned char)(bgindex & 0xff);
      image_fill_pixel(image, work->type, &bgi);
    } else {
      image_fill_pixel(image, work->type, bg);
    }
    full_image = TRUE;
    break;
  case DISPOSE_OP_PREVIOUS:
    image_copy_data(image, work->previous, work->type);
    full_image = TRUE;
    break;
  default:
    DEBUG_PRINT(" unknown dipose_op type %d. why?\n", dispose_op0);
    break;
  }
  if (full_image) {
    image_get_size(image, &w, &h);
    xoff = 0;
    yoff = 0;
  } else {
    image_get_size(newframe, &w, &h);
    image_get_offset(newframe, &xoff, &yoff);
  }
  work->updated_xoff = xoff;
  work->updated_yoff = yoff;
  work->updated_width = w;
  work->updated_height = h;
  if (work->type == IMAGE_TYPE_INDEXED) {
    work->dispose_op = dispose_op0;
  } else {
    work->dispose_op = DISPOSE_OP_NONE;
  }
  image_set_dispose_op(image, dispose_op);
  work->previous = last_frame;
  if (work->type == IMAGE_TYPE_INDEXED) {
    CmsProfile *profile_in = imageWork_get_profile_in(work);
    CmsProfile *profile_out = work->profile_out;
    if ((profile_in != NULL) && (profile_out != NULL)) {
      image_transform_pallete(newframe, profile_in, image, profile_out);
    } else {
      image_copy_pallete(image, newframe);
    }
  }
  /* blend operation */
  blend_op = image_get_blend_op(newframe);
  DEBUG_PRINT(" blend_op type: %d\n", blend_op);
  switch (blend_op) {
  case BLEND_OP_NONE:
    image_blend(work, newframe, newtype, bg, bgindex);
    break;
  case BLEND_OP_SOURCE:
    image_blend(work, newframe, newtype, bg, bgindex);
    break;
  case BLEND_OP_OVER:
    image_blend(work, newframe, newtype, NULL, -1);
    break;
  default:
    DEBUG_PRINT(" unknown blend_op type %d. why?\n", blend_op);
    break;
  }
  if (work->type != IMAGE_TYPE_INDEXED) {
    blend_op = BLEND_OP_NONE;
  }
  image_set_blend_op(image, blend_op);
  image_set_type(image, work->type);
}

int imageWork_get_dispose_op(ImageWork *work)
{
  return work->dispose_op;
}

Image *imageWork_get_image(ImageWork *work)
{
  return work->current;
}

void imageWork_get_updated_size(ImageWork *work, int *width, int *height, int *xoff, int *yoff)
{
  *xoff = work->updated_xoff;
  *yoff = work->updated_yoff;
  *width = work->updated_width;
  *height = work->updated_height;
}

void imageWork_set_profiles(ImageWork *work, CmsProfile *profile_in, CmsProfile *profile_out, int use_infile)
{
  work->profile_in = profile_in;
  work->profile_out = profile_out;
  work->use_infile = use_infile;
}

static CmsProfile *imageWork_get_profile_in(ImageWork *work)
{
  CmsProfile *profile;

  if (work->use_infile) {
    profile = work->profile_in;
  } else {
    profile = work->profile_embedded;
    if (profile == NULL) {
      profile = work->profile_in;
    }
  }
  return profile;
}

static void image_copy_data(Image *dst, Image *src, ImageType type)
{
  unsigned char *pp;
  unsigned char *pdst;
  unsigned char *psrc;
  int w0, h0, w, h;
  int xoff, yoff;
  size_t bpp0, bpp; /* bytes per pixel */

  pdst = image_get_data(dst);
  image_get_size(dst, &w0, &h0);
  psrc = image_get_data(src);
  image_get_size(src, &w, &h);
  image_get_offset(src, &xoff, &yoff);
  bpp = bpp0 = imageType_bytes_per_pixel(type);
  pp = pdst+(w0*yoff+xoff)*bpp0;
  if (w == w0) {
    memcpy(pp, psrc, w*h*bpp);
  } else {
    unsigned char *pp2;
    pp2 = psrc;
    for (int j = 0; j < h; j++) {
      memcpy(pp, pp2, w*bpp);
      pp += w0*bpp0;
      pp2 += w*bpp;
    }
  }
}

static void image_copy_pallete(Image *dst, Image *src)
{
  int num;
  num = image_get_num_colors(src);
  image_set_num_colors(dst, num);
  for (int i = 0; i < num; i++) {
    unsigned char r, g, b, a;
    image_get_color(src, i, &r, &g, &b, &a);
    image_set_color(dst, i, r, g, b, a);
  }
}

static void image_fill_pixel(Image *image, ImageType type, unsigned char *pixel)
{
  unsigned char *pp;
  int w, h, bpp = -1;

  image_get_size(image, &w, &h);
  pp = image_get_data(image);
  switch (type) {
  case IMAGE_TYPE_INDEXED:
    for (int j = 0; j < h; j++) {
      for (int i = 0; i < w; i++) {
        *(pp++) = *pixel;
      }
    }
    break;
  case IMAGE_TYPE_RGB_8:
  case IMAGE_TYPE_RGBA_8:
  case IMAGE_TYPE_RGB_16:
  case IMAGE_TYPE_RGBA_16:
    bpp = imageType_bytes_per_pixel(type);
    for (int j = 0; j < h; j++) {
      for (int i = 0; i < w; i++) {
        for (int c = 0; c < bpp; c++) {
          *(pp++) = pixel[c];
        }
      }
    }
    break;
  }
}

static void image_blend(ImageWork *work, Image *newframe, ImageType type,
                        const unsigned char *bg, int bgindex)
{
  Image *image;
  unsigned char *pp;
  size_t bpp0, bpp; /* bytes per pixel */
  unsigned char *pp1;
  unsigned char *pp2;
  int w0, h0, w, h;
  int xoff, yoff;
  CmsProfile *profile_in = NULL;
  CmsProfile *profile_out = NULL;

  image = work->current;
  pp1 = image_get_data(image);
  image_get_size(image, &w0, &h0);
  pp2 = image_get_data(newframe);
  image_get_size(newframe, &w, &h);
  image_get_offset(newframe, &xoff, &yoff);
  switch (type) {
  case IMAGE_TYPE_INDEXED:
    if (work->type == IMAGE_TYPE_RGB_8) {
      /* +indexed color -> RGB8 */
      image_blend_index_to_rgb8(image, newframe, bg, bgindex);
    } else {
      /* +indexed color -> indexed color */
      image_copy_data(image, newframe, type);
    }
    break;
  case IMAGE_TYPE_RGB_8:
  case IMAGE_TYPE_RGB_16:
    /* +RGB -> RGB */
    profile_in = imageWork_get_profile_in(work);
    profile_out = work->profile_out;
    if ((profile_in != NULL) && (profile_out != NULL)) {
      image_transform_data(newframe, profile_in, image, profile_out);
    } else {
      image_copy_data(image, newframe, type);
    }
    break;
  case IMAGE_TYPE_RGBA_8:
  case IMAGE_TYPE_RGBA_16:
    /* +RGBA -> RGB */
    profile_in = imageWork_get_profile_in(work);
    profile_out = work->profile_out;
    if ((profile_in != NULL) && (profile_out != NULL)) {
      Image *tmpframe;
      tmpframe = image_transform_new(newframe, profile_in, profile_out);
      image_blend_rgb(image, tmpframe, type, bg);
      image_free(tmpframe);
    } else {
      image_blend_rgb(image, newframe, type, bg);
    }
    break;
  default:
    break;
  }
}

/*
 * +indexed color -> RGB8
 */
static void image_blend_index_to_rgb8(Image *image, Image *newframe, const unsigned char *bg, int bgindex)
{
  unsigned char *pp;
  unsigned char *pp1;
  unsigned char *pp2;
  unsigned char bgc[4];
  int w0, h0, w, h, xoff, yoff;
  int bpp0;

  bpp0 = imageType_bytes_per_pixel(IMAGE_TYPE_RGB_8);
  if (bg == NULL) {
    image_get_color(newframe, bgindex, &bgc[0], &bgc[1], &bgc[2], &bgc[3]);
    bg = bgc;
  }
  pp1 = image_get_data(image);
  image_get_size(image, &w0, &h0);
  pp2 = image_get_data(newframe);
  image_get_size(newframe, &w, &h);
  image_get_offset(newframe, &xoff, &yoff);
  for (int j = 0; j < h; j++) {
    pp = pp1+(w0*(yoff+j)+xoff)*bpp0;
    for (int i = 0; i < w; i++) {
      unsigned char rgba[4];
      image_get_color(newframe, *pp2, &rgba[0], &rgba[1], &rgba[2], &rgba[3]);
      blend_rgb8(pp, rgba, bg);
      pp += bpp0;
      pp2++;
    }
  }
}

/*
 * +RGBA -> RGB
 */
static void image_blend_rgb(Image *image, Image *newframe, ImageType type, const unsigned char *bg)
{
  void (*blend_rgb)(unsigned char *p1, const unsigned char *p2, const unsigned char *bg);
  ImageType type0 = IMAGE_TYPE_UNKNOWN;
  size_t bpp0, bpp; /* bytes per pixel */
  unsigned char *pp1;
  unsigned char *pp2;
  int w0, h0, w, h;
  int xoff, yoff;

  switch (type) {
  case IMAGE_TYPE_RGBA_8:
    type0 = IMAGE_TYPE_RGB_8;
    blend_rgb = blend_rgb8;
    break;
  case IMAGE_TYPE_RGBA_16:
    type0 = IMAGE_TYPE_RGB_16;
    blend_rgb = blend_rgb16;
    break;
  default:
    return;
    break;
  }
  bpp0 = imageType_bytes_per_pixel(type0);
  bpp = imageType_bytes_per_pixel(type);
  pp1 = image_get_data(image);
  image_get_size(image, &w0, &h0);
  pp2 = image_get_data(newframe);
  image_get_size(newframe, &w, &h);
  image_get_offset(newframe, &xoff, &yoff);
  for (int j = 0; j < h; j++) {
    unsigned char *pp;
    pp = pp1+(w0*(yoff+j)+xoff)*bpp0;
    for (int i = 0; i < w; i++) {
      (*blend_rgb)(pp, pp2, bg);
      pp += bpp0;
      pp2 += bpp;
    }
  }
}

static void blend_rgb8(unsigned char *p1, const unsigned char *p2, const unsigned char *bg)
{
  unsigned char *pp;
  unsigned char a;
  int bpp; /* bytes_per_pixel */

  bpp = sizeof(unsigned char)*1*3;
  a = *(p2+3);
  if (a == 0xff) {
    /* fully opaque */
    memcpy(p1, p2, bpp);
    return;
  }
  if (bg != NULL) {
    /* case BLEND_OP_SOURCE */
    if (a == 0x00) {
      /* fully transparent */
      memcpy(p1, bg, bpp);
      return;
    }
    if (memcmp(bg, p2, bpp) == 0) {
      /* same value */
      memcpy(p1, p2, bpp);
      return;
    }
    pp = p1;
    p1 = bg;
  } else {
    /* case BLEND_OP_OVER */
    if (a == 0x00) {
      /* fully transparent */
      return;
    }
    if (memcmp(p1, p2, bpp) == 0) {
      /* same value */
      return;
    }
    pp = p1;
  }
  for (int i = 0; i < 3; i++) {
    unsigned int c1, c2, cc;
    unsigned char c;

    c1 = *(p1++);
    c2 = *(p2++);
    cc = (c1*(0xff-a)+c2*a)/0xff; /* = c1*(1.0-alpha)+c2*alpha */
    c = (unsigned char)cc;
    *(pp++) = c;
  }
}

static void blend_rgb16(unsigned char *p1, const unsigned char *p2, const unsigned char *bg)
{
  unsigned char *pp;
  unsigned short a;
  int bpp; /* bytes_per_pixel */

  bpp = sizeof(unsigned char)*2*3;
  a = get_c16(p2+6);
  if (a == 0xffff) {
    /* fully opaque */
    memcpy(p1, p2, bpp);
    return;
  }
  if (bg != NULL) {
    /* case BLEND_OP_SOURCE */
    if (a == 0x0000) {
      /* fully transparent */
      memcpy(p1, bg, bpp);
      return;
    }
    if (memcmp(bg, p2, bpp) == 0) {
      /* same value */
      memcpy(p1, p2, bpp);
      return;
    }
    pp = p1;
    p1 = bg;
  } else {
    /* case BLEND_OP_OVER */
    if (a == 0x0000) {
      /* fully transparent */
      return;
    }
    if (memcmp(p1, p2, bpp) == 0) {
      /* same value */
      return;
    }
    pp = p1;
  }
  for (int i = 0; i < 3; i++) {
    unsigned long c1, c2, cc;
    unsigned short c;

    c1 = get_c16(p1);
    c2 = get_c16(p2);
    cc = (c1*(0xffff-a)+c2*a)/0xffff; /* = c1*(1.0-alpha)+c2*a */
    c = (unsigned short)cc;
    set_c16(pp, c);
    p1 += 2;
    p2 += 2;
    pp += 2;
  }
}

static void get_rgb8(unsigned char *rgb, unsigned short *r, unsigned short *g, unsigned short *b)
{
  *r = get_c8(&rgb[0]);
  *g = get_c8(&rgb[1]);
  *b = get_c8(&rgb[2]);
}

static void set_rgb8(unsigned char *rgb, unsigned short r, unsigned short g, unsigned short b)
{
  set_c8(rgb, r);
  set_c8(rgb+1, g);
  set_c8(rgb+2, b);
}

static void get_rgb16(unsigned char *rgb, unsigned short *r, unsigned short *g, unsigned short *b)
{
  *r = get_c16(&rgb[0]);
  *g = get_c16(&rgb[2]);
  *b = get_c16(&rgb[4]);
}

static void set_rgb16(unsigned char *rgb, unsigned short r, unsigned short g, unsigned short b)
{
  set_c16(&rgb[0], r);
  set_c16(&rgb[2], g);
  set_c16(&rgb[4], b);
}

static unsigned short get_c16(unsigned char *c16)
{
  unsigned short c;

  c  = (c16[0] << 8) & 0xff00;
  c |= c16[1] & 0xff;
  return c;
}

static void set_c16(unsigned char *c16, unsigned short c)
{
  c16[0] = (c >> 8) & 0xff;
  c16[1] = c & 0xff;
}

static unsigned short get_c8(unsigned char *c8)
{
  unsigned short cc, c;

  cc = c8[0];
  c  = (cc << 8) & 0xff00;
  c |= cc & 0xff;
  return c;
}

static void set_c8(unsigned char *c8, unsigned short c)
{
  *c8 = (c >> 8) & 0xff;
}
