/* $Id$ */
#include <math.h>
#include "common.h"
#include "module.h"
#include "cmsprofile.h"

struct backgroundColor {
  unsigned short red;
  unsigned short green;
  unsigned short blue;
  unsigned char index;
};
typedef struct backgroundColor BackgroundColor;

struct imageColor {
  unsigned char red;
  unsigned char green;
  unsigned char blue;
  unsigned char alpha;
};
typedef struct imageColor ImageColor;

struct image {
  unsigned char *data; /* image data */
  int xoff, yoff;      /* offset for image */
  int width, height;   /* image size */
  ImageType type;      /* type of image */
  int num_colors;      /* number of indexed colors */
  ImageColor *colors;  /* color map for indexed palette */
  int delay;           /* interframe delay [milli second] */
  int dispose_op;      /* dispose operation */
  int blend_op;        /* blend operation */
  const char *comment; /* comment text */
  struct image *next;  /* pointer to next image */
};
struct imageSet {
  Image *frames;
  int width, height;   /* images dimensions */
  int num_images;      /* number of images */
  int num_loop;        /* number of loop */
  int num_colors;      /* number of indexed colors */
  ImageColor *colors;  /* color map for indexed palette */
  BackgroundColor *background;
  CmsProfile *profile; /* color profile */
};
struct reader {
  const char *key;    /* magic data */
  size_t key_len;     /* magic data length */
  const char *format; /* image format */
  int (*mod_read)(Entry *entry, ImageSet *); /* read function */
};
typedef struct reader Reader;

#if defined ENABLE_GIF
extern int mod_read_GIF(Entry *entry, ImageSet *images);
#endif /* ENABLE_GIF */
#if defined ENABLE_JPEG
extern int mod_read_JPEG(Entry *entry, ImageSet *images);
#endif /* ENABLE_JPEG */
#if defined ENABLE_PNG
extern int mod_read_PNG(Entry *entry, ImageSet *images);
#endif /* ENABLE_PNG */
#if defined ENABLE_TIFF
extern int mod_read_TIFF(Entry *entry, ImageSet *images);
#endif /* ENABLE_TIFF */
extern int mod_read_xbm(Entry *entry, ImageSet *images);
extern int mod_read_null(Entry *entry, ImageSet *images);

static void imageSet_init(ImageSet *images);
static void imageColor_get_color(ImageColor *color,unsigned char *r, unsigned char *g, unsigned char *b, unsigned char *a);
static void imageColor_set_color(ImageColor *color, unsigned char r, unsigned char g, unsigned char b, unsigned char a);


int module_read_images(Entry *entry, ImageSet *images, const char *format)
{
  const Reader readers[] = {
#if defined ENABLE_GIF
    { "GIF",      3, "gif", mod_read_GIF },
#endif /* ENABLE_GIF */
#if defined ENABLE_JPEG
    { "\xff\xd8", 2, "jpg", mod_read_JPEG },
#endif /* ENABLE_JPEG */
#if defined ENABLE_PNG
    { "\x89PNG",  4, "png", mod_read_PNG },
#endif /* ENABLE_PNG */
#if defined ENABLE_TIFF
    { "MM\x00\x2a", 4, "tif", mod_read_TIFF },
    { "II\x2a\x00", 4, "tif", mod_read_TIFF },
#endif /* ENABLE_TIFF */
    { "#define", 7, "xbm", mod_read_xbm },
    { NULL, 0, NULL, mod_read_null }
  };
  const Reader *reader;

  if (format == NULL) {
    FILE *fp;
    char buf[16];
    fp = fopen(entry->filename, "rb");
    if (fp == NULL) {
      return FALSE;
    }
    if (fread(buf, sizeof(char), 16, fp) != sizeof(char)*16) {
      return FALSE;
    }
    fclose(fp);
    reader = readers;
    while (reader->key != NULL) {
      if (memcmp(buf, reader->key, reader->key_len) == 0) {
        break;
      }
      reader++;
    }
    if (reader->key == NULL) {
      fprintf(stderr, "Unknown file type '%s'\n", entry->filename);
    }
  } else {
    reader = readers;
    while (reader->key != NULL) {
      if (strcmp(format, reader->format) == 0) {
        break;
      }
      reader++;
    }
    if (reader->key == NULL) {
      fprintf(stderr, "Unknown format type '%s'\n", format);
    }
  }
  imageSet_init(images);
  if (!(*(reader->mod_read))(entry, images)) {
    fprintf(stderr, "Can not read file '%s'\n", entry->filename);
    return FALSE;
  }
  return TRUE;
}

ImageSet *imageSet_new(void)
{
  ImageSet *images;

  images = XMALLOC(ImageSet, 1);
  if (images != NULL) {
    imageSet_init(images);
  }
  return images;
}

void imageSet_free(ImageSet *images)
{
  Image *frame;
  int n;

  n = images->num_images;
  frame = images->frames;
  for (int i = 0; i < n; i++) {
    Image *next;
    next = image_next(frame);
    image_free(frame);
    frame = next;
    DEBUG_PRINT("imageSet_free: %d\n", i);
  }
  images->frames = NULL;
  images->num_images = 0;
  images->num_loop = 0;
  images->num_colors = 0;
  XFREE(images->colors);
  XFREE(images->background);
  cmsProfile_close(images->profile);
}

/*
 * add an image frame.
 */
void imageSet_add_image(ImageSet *images, Image *image)
{
  Image *p;

  p = images->frames;
  if (p == NULL) {
    images->frames = image;
    images->num_images++;
    return;
  }
  while (p->next != NULL) {
    p = p->next;
  }
  p->next = image;
  images->num_images++;
}

/*
 * gets number of images.
 */
int imageSet_get_num_frames(ImageSet *images)
{
  return images->num_images;
}

/*
 * gets first image.
 */
Image *imageSet_get_first(ImageSet *images)
{
  return images->frames;
}

/*
 * gets background color index.
 */
unsigned char imageSet_get_background_index(ImageSet *images)
{
  BackgroundColor *color;

  color = images->background;
  return color->index;
}

/*
 * sets background color index.
 */
void imageSet_set_background_index(ImageSet *images, unsigned char index)
{
  BackgroundColor *color;

  color = images->background;
  color->index = index;
}

/*
 * gets background color RGB.
 */
void imageSet_get_background_rgb(ImageSet *images, unsigned short *r, unsigned short *g, unsigned short *b)
{
  BackgroundColor *color;

  color = images->background;
  *r = color->red;
  *g = color->green;
  *b = color->blue;
}

/*
 * sets background color RGB.
 */
void imageSet_set_background_rgb(ImageSet *images, unsigned short r, unsigned short g, unsigned short b)
{
  BackgroundColor *color;

  color = images->background;
  color->red   = r;
  color->green = g;
  color->blue  = b;
}

/*
 * gets screen size.
 */
void imageSet_get_size(ImageSet *images, int *width, int *height)
{
  *width = images->width;
  *height = images->height;
}

/*
 * sets screen size.
 */
void imageSet_set_size(ImageSet *images, int width, int height)
{
  images->width = width;
  images->height = height;
}

/*
 * gets number of loop.
 */
int imageSet_get_num_plays(ImageSet *images)
{
  return images->num_loop;
}

/*
 * sets number of loop.
 */
void imageSet_set_num_plays(ImageSet *images, int num)
{
  images->num_loop = num;
}

/*
 * gets number of index colors.
 */
int imageSet_get_num_colors(ImageSet *images)
{
  return images->num_colors;
}

/*
 * sets number of index colors.
 */
void imageSet_set_num_colors(ImageSet *images, int num)
{
  if (num > 0) {
    XFREE(images->colors);
    images->num_colors = num;
    images->colors = XMALLOC(ImageColor, num);
  }
}

/*
 * gets indexed color.
 */
void imageSet_get_color(ImageSet *images, int index, unsigned char *r, unsigned char *g, unsigned char *b, unsigned char *a)
{
  ImageColor *colors;

  colors = images->colors;
  if (index < images->num_colors) {
    imageColor_get_color(&colors[index], r, g, b, a);
  }
}

/*
 * sets indexed color.
 */
void imageSet_set_color(ImageSet *images, int index, unsigned char r, unsigned char g, unsigned char b, unsigned char a)
{
  ImageColor *colors;

  colors = images->colors;
  if (index < images->num_colors) {
    imageColor_set_color(&colors[index], r, g, b, a);
  }
}

/*
 * gets image color profile.
 */
CmsProfile *imageSet_get_profile(ImageSet *images)
{
  return images->profile;
}

/*
 * sets image color profile.
 */
void imageSet_set_profile(ImageSet *images, CmsProfile *profile)
{
  images->profile = profile;
}

static void imageSet_init(ImageSet *images)
{
  BackgroundColor *bg;

  images->frames = NULL;
  images->width = 320;
  images->height = 200;
  images->num_images = 0;
  images->num_loop = 0;
  images->num_colors = 0;
  images->colors = NULL;
  bg = XMALLOC(BackgroundColor, 1);
  bg->red = 0x7f7f;
  bg->green = 0x7f7f;
  bg->blue = 0x7f7f;
  bg->index = 0x7f;
  images->background = bg;
  images->profile = NULL;
}

Image *image_new(void)
{
  Image *image;

  image = XMALLOC(Image, 1);
  if (image == NULL) {
    return NULL;
  }
  image->data = NULL;
  image->xoff = 0;
  image->yoff = 0;
  image->width = 0;
  image->height = 0;
  image->type = IMAGE_TYPE_UNKNOWN;
  image->num_colors = 0;
  image->colors = NULL;
  image->delay = 0;
  image->dispose_op = DISPOSE_OP_NONE;
  image->blend_op = BLEND_OP_NONE;
  image->comment = NULL;
  image->next = NULL;
  return image;
}

void image_free(Image *image)
{
  unsigned char *data;
  ImageColor *colors;
  char *comment;

  data = image->data;
  colors = image->colors;
  comment = (char *)image->comment;
  XFREE(data);
  XFREE(colors);
  XFREE(comment);
  XFREE(image);
}

/*
 * gets image data.
 */
unsigned char *image_get_data(Image *image)
{
  return image->data;
}

/*
 * sets image data.
 */
void image_set_data(Image *image, unsigned char *data)
{
  image->data = data;
}

/*
 * gets image size.
 */
void image_get_size(Image *image, int *width, int *height)
{
  *width = image->width;
  *height = image->height;
}

/*
 * sets image size.
 */
void image_set_size(Image *image, int width, int height)
{
  image->width = width;
  image->height = height;
}

/*
 * gets image offset.
 */
void image_get_offset(Image *image, int *xoff, int *yoff)
{
  *xoff = image->xoff;
  *yoff = image->yoff;
}

/*
 * sets image offset.
 */
void image_set_offset(Image *image, int xoff, int yoff)
{
  image->xoff = xoff;
  image->yoff = yoff;
}

/*
 * gets number of index colors.
 */
int image_get_num_colors(Image *image)
{
  return image->num_colors;
}

/*
 * sets number of index colors.
 */
void image_set_num_colors(Image *image, int num)
{
  if (num > 0) {
    XFREE(image->colors);
    image->num_colors = num;
    image->colors = XMALLOC(ImageColor, num);
  }
}

/*
 * gets indexed color.
 */
void image_get_color(Image *image, int index, unsigned char *r, unsigned char *g, unsigned char *b, unsigned char *a)
{
  ImageColor *colors;

  colors = image->colors;
  if (index < image->num_colors) {
    imageColor_get_color(&colors[index], r, g, b, a);
  }
}

/*
 * sets indexed color.
 */
void image_set_color(Image *image, int index, unsigned char r, unsigned char g, unsigned char b, unsigned char a)
{
  ImageColor *colors;

  colors = image->colors;
  if (index < image->num_colors) {
    imageColor_set_color(&colors[index], r, g, b, a);
  }
}

/*
 * sets comment.
 */
void image_set_comment(Image *image, const char *comment)
{
  image->comment = comment;
}

/*
 * gets delay [m sec].
 */
int image_get_delay(Image *image)
{
  return image->delay;
}

/*
 * sets delay [m sec].
 */
void image_set_delay(Image *image, int delay)
{
  image->delay = delay;
}

/*
 * gets dispose operation.
 */
int image_get_dispose_op(Image *image)
{
  return image->dispose_op;
}

/*
 * sets dispose operation.
 */
void image_set_dispose_op(Image *image, int dispose_op)
{
  image->dispose_op = dispose_op;
}

/*
 * gets blend operation.
 */
int image_get_blend_op(Image *image)
{
  return image->blend_op;
}

/*
 * sets blend operation.
 */
void image_set_blend_op(Image *image, int blend_op)
{
  image->blend_op = blend_op;
}

/*
 * gets image type.
 */
ImageType image_get_type(Image *image)
{
  return image->type;
}

/*
 * sets image type.
 */
void image_set_type(Image *image, ImageType type)
{
  image->type = type;
}

/*
 * next image.
 */
Image *image_next(Image *image)
{
  return image->next;
}

/*
 * size of image type.
 */
int imageType_bytes_per_pixel(ImageType type)
{
  int bpp = -1;

  switch (type) {
  case IMAGE_TYPE_INDEXED:
    bpp = sizeof(unsigned char);
    break;
  case IMAGE_TYPE_RGB_8:
    bpp = sizeof(unsigned char)*3;
    break;
  case IMAGE_TYPE_RGBA_8:
    bpp = sizeof(unsigned char)*4;
    break;
  case IMAGE_TYPE_RGB_16:
    bpp = sizeof(unsigned char)*6;
    break;
  case IMAGE_TYPE_RGBA_16:
    bpp = sizeof(unsigned char)*8;
    break;
  default:
    break;
  }
  return bpp;
}

static void imageColor_get_color(ImageColor *color,unsigned char *r, unsigned char *g, unsigned char *b, unsigned char *a)
{
  *r = color->red;
  *g = color->green;
  *b = color->blue;
  *a = color->alpha;
}

static void imageColor_set_color(ImageColor *color, unsigned char r, unsigned char g, unsigned char b, unsigned char a)
{
  color->red = r;
  color->green = g;
  color->blue = b;
  color->alpha = a;
}
